package Gtk2::Ex::HyperTextBuffer;

our $VERSION = '0.15';

use strict;
use Gtk2;
use Gtk2::Pango; # pango constants
use File::Spec;  # for locating images
use Glib::Object::Subclass Gtk2::TextBuffer:: ;

=head1 NAME

HyperTextBuffer - A TextBuffer for the HyperTextView widget

=head1 DESCRIPTION

This module is a subclass of L<Gtk2::TextBuffer> intended for use with 
L<Gtk2::Ex::HyperTextView>. It has a high level interface that uses parse trees
instead of plain text. It also keeps an undo-/redo-stack.

=head1 HIERARCHY

  Glib::Object
  +----Gtk2::TextBuffer
       +----Gtk2::Ex::HyperTextBuffer


=cut

# It seems to me that 10 * PANGO_SCALE is the normal font size
# TODO make tag styles configable
# FIXME replace "weight" with "scale" ?
our %TAGS = (
	bold      => [weight => PANGO_WEIGHT_BOLD],
	italic    => [style => 'italic'],
	underline => [underline => 'single'],
	strike    => [strikethrough => 'true', foreground => 'grey'],
	verbatim  => [family => 'monospace'], # wrap_mode => 'none'], # Bug withwrap_mode !?
	Verbatim  => [family => 'monospace', wrap_mode => 'none', indent => 40],
	link      => [foreground => 'blue', underline => 'single'],
	head1     => [weight => PANGO_WEIGHT_BOLD, size => 10*1.2**3 * PANGO_SCALE ],
	head2     => [weight => PANGO_WEIGHT_BOLD, size => 10*1.2**2 * PANGO_SCALE ],
	head3     => [weight => PANGO_WEIGHT_BOLD, size => 12 * PANGO_SCALE        ],
	head4     => [weight => PANGO_WEIGHT_BOLD, size => 10 * PANGO_SCALE        ],
	head5     => [style  => 'italic', weight => PANGO_WEIGHT_BOLD, size => 10 * PANGO_SCALE],
);

our @_tags = grep {$_ ne 'link'} keys %TAGS;

our ($_clipboard, $_primary);

our @_targets = (
	{ target => 'STRING'        },
	{ target => 'TEXT'          },
	{ target => 'COMPOUND_TEXT' },
	{ target => 'UTF8_STRING'   },
	{ target => 'ZIM_TREE'      },
);

=head1 METHODS

=over 4

=item C<< new(PROPERTY => VALUE, ..) >>

Simple constructor.

=cut

sub INIT_INSTANCE {
	my $self = shift;
	$self->{edit_mode_tags} = [];
	$self->{image_marks} = [];
#	$self->signal_connect(mark_set => sub {
#			my ($self, $iter, $mark) = @_;
#			return unless $mark->get_name eq 'insert';
#			my @names = map {$_->get_property('name')} $self->_get_open_tags($iter);
#			warn "Tags: @names\n";
#			#$self->set_edit_mode_from_cursor;
#		} );
}

=item C<clear()>

Deletes all buffer contents.

Be aware that this module sometimes stores a reference to the buffer in a
global variable in order to get copy-pasting with mark up work. This means
you could save a lot of memory when you always clear the object when you are
done with it.

=cut

sub clear {
	my $self = shift;
	$self->{edit_mode_tags} = [];
	$self->{image_marks} = [];
	$self->delete( $self->get_bounds );
}

=item C<create_default_tags()>

Initialise pre-defined text tags. See TAGS for a listing of these tags.

=cut

sub create_default_tags {
	$_[0]->create_tag($_ => @{$TAGS{$_}}) for @_tags;
}

=item C<set_parse_tree(TREE)>

Load the buffer with contents from a parse tree. The tree format is based on
the format as used by L<Pod::Simple> but it uses more verbose element names.

For example:

	['Document', {},
		['head1', {}, 'Test page'],
		"\n\n",
		"This is a test page showing some syntax:\n\n",
		['bold', {}, 'Bold text'], "\n",
		['italic', {}, 'Italic text'], "\n",
		['bold', {},
			['italic', {}, 'Bold and Italic']
		], "\n",
	]

=cut

sub set_parse_tree {
	my ($self, $tree) = @_;
	#use Data::Dumper; print "HyperTextBuffer got tree: ", Dumper $tree;
	$self->clear;
	$self->insert_blocks_at_cursor($tree);
	$self->set_modified(0);
}

=item C<insert_blocks(ITER, BLOCK, ...)>

Insert one or more partial parse trees in the buffer at ITER.

=cut

sub insert_blocks {
	my $self = shift;
	my $iter = shift;
	$self->place_cursor($iter);
	$self->set_edit_mode_from_cursor;
	$self->insert_blocks_at_cursor(@_);
}

=item C<insert_blocks_at_cursor(BLOCK, ...)>

Like C<insert_blocks()> but inserts at current cursor position.

=cut

my $_Verbatim = 0;

sub insert_blocks_at_cursor {
	# Make sure not to modify content of the reference !
	# TODO: check default editable / region editable
	my ($self, @blocks) = @_;
#	$self->set_edit_mode_from_cursor
#		unless defined $self->{edit_mode_tags}; # FIXME
	$self->{edit_mode_tags} ||= [];
	for (@blocks) {
		unless (ref $_) {
			$_ =~ s/^(\s*)\*(\s+)/$1\x{2022}$2/mg unless $_Verbatim; # FIXME doesn't belong in this module !
			if (@{$self->{edit_mode_tags}}) {
				my $iter = $self->get_iter_at_mark( $self->get_insert );
				$self->insert_with_tags(
					$iter, $_, @{$self->{edit_mode_tags}} );
			}
			else { $self->insert_at_cursor($_) }
		}
		else {
			my ($type, $meta) = @$_[0, 1];
			my $tag;
			if ($type eq 'image') {
				$self->insert_image_from_file($$meta{src});
				next;
			}
			elsif ($type eq 'link') {
				my $bit = (@$_ == 3 and $$_[2] eq $$meta{to});
				$tag = $self->create_link_tag($bit ? undef : $$meta{to});
			}
			else { $tag = $self->get_tag_table->lookup($type) }
			
			if ($tag) {
				$_Verbatim++ if $type eq 'Verbatim'; # FIXME
				push @{$self->{edit_mode_tags}}, $tag;
				$self->insert_blocks_at_cursor(@$_[2 .. $#$_]); # recurs
				pop @{$self->{edit_mode_tags}};
				$_Verbatim-- if $type eq 'Verbatim'; # FIXME
			}
			elsif ($type eq 'Document' or $type eq 'Para') {
				if ($$meta{base}) {
					$self->{base} = $$meta{base};
					$self->{base} =~ s#^file://##;
					#warn "base set to: $self->{base}\n";
				}
				$self->insert_blocks_at_cursor(@$_[2 .. $#$_]); # recurs
			}
			else { die "Unknown tag in parse tree: $type" }

			my $lines = ($type =~ /^head/) ? 1 : 0;
			$lines += $$meta{empty_lines} if $$meta{empty_lines}; 
			$self->insert_at_cursor("\n"x$lines) if $lines;
		}
	}

	return 1;
}

=item C<insert_image_from_file(FILENAME)>

Insert an image in the buffer at the cursor.

=cut

sub insert_image_from_file {
	my ($self, $src) = @_;
	my $file = $self->{base} ? File::Spec->rel2abs($src, $self->{base}) : $src;
	#print "file: $$meta{src}, base: $self->{base} => $file\n";

	my $pixbuf = Gtk2::Gdk::Pixbuf->new_from_file($file) if -f $file and -r _;

	$pixbuf ||= Gtk2::Image->new->render_icon('gtk-missing-image', 'dialog')->copy;
	die "could not insert image: $file\n" unless $pixbuf;
	
	if (defined $pixbuf->{image_src}) {
		# copy does not always copy, or so it seems :(
		$pixbuf = $pixbuf->copy;
		warn "WARNING: pixbuf bug !??";
	}
	die "ERROR: Pixbuf BUG !" if defined $pixbuf->{image_src};
	#warn "PixBuf: $pixbuf ($pixbuf->{image_src})\n";
	
	$pixbuf->{image_src} = $src;
	$pixbuf->{image_file} = $file;
	my $iter = $self->get_iter_at_mark( $self->get_insert );

	#my $mark = $self->create_mark(undef, $iter, 0);
	#$mark->set_visible(1);
	
	$self->insert_pixbuf($iter, $pixbuf);

	return 1;
}

=item C<get_parse_tree()>

Returns a parse tree based on the buffer contents.

=cut

sub get_parse_tree {
	# TODO Nesting order of tags starting on the same
	# position is arbitrary - add logic to sort this out
	# TODO: restructure the tree to order it in paragraphs
	my ($self, $start, $end) = @_;
	($start, $end) = $self->get_bounds unless defined $start;

	# First we build a parse tree based on the tags we find
	my $root = [];
	my @stack = ($root);
	my $iter = $start->copy;
	for ($iter->get_tags) {
		my $link_data = Gtk2::Ex::HyperTextView->get_link_at_iter($iter, $_)
			if $_->{is_link}; # FIXME quick hack here
		my $tag = $_->{is_link}
			? [ 'link', {to => $link_data}   ]
			: [ $_->get_property('name'), {} ] ;
		next if $$tag[0] =~ /gtkspell/;
		push @{$stack[-1]}, $tag;
		push @stack, $tag;
	}
	my $previous = $iter->copy;
	while ($iter->forward_to_tag_toggle(undef)) {
		last if $iter->compare($end) != -1;
		
		EXTRACT_TAGS:
		my $slice = $previous->get_slice($iter);
		$slice =~ s/^(\s*)\x{2022}(\s+)/$1\*$2/mg; # FIXME totally wrong level
		if ($slice =~ /\x{fffc}/) {
			my $image = $previous->copy;
			while ($slice =~ s/^(.*?)\x{fffc}//s) { # Embeded image
				push @{$stack[-1]}, $1 if length $1;
				$image->forward_chars(length $1);
				my $pixbuf = $image->get_pixbuf;
				if ($pixbuf) {
					my $src = $pixbuf->{image_src};
					push @{$stack[-1]}, ['image', {src => $src}];
				}
				$image->forward_char;
			}
		}
		push @{$stack[-1]}, $slice if length $slice;
		$previous = $iter->copy;
		for ($iter->get_toggled_tags(0)) { # closed tags
			# warning: no consistency check is being done here
			#          because order of closing tags is unsorted
			next if !$_->{is_link}
				and $_->get_property('name') =~ /gtkspell/;
			pop @stack;
		}
		for ($iter->get_toggled_tags(1)) { # opened tags
			my $link_data = Gtk2::Ex::HyperTextView->get_link_at_iter($iter, $_)
				if $_->{is_link}; # FIXME quick hack here
			my $tag = $_->{is_link}
				? [ 'link', {to => $link_data}   ]
				: [ $_->get_property('name'), {} ] ;
			next if $$tag[0] =~ /gtkspell/;
			push @{$stack[-1]}, $tag;
			push @stack, $tag;
		}
	}
	unless ($previous->equal($end)) { # last part
		$iter = $end->copy;
		goto EXTRACT_TAGS;
	}

	# Check lines around headings - FIXME very ugly implementation
	# Also add sanity check to merge blocks of 'verbatim' to 'Verbatim'
	for (0 .. $#$root) {
		next unless ref $$root[$_] and $$root[$_][0] =~ /^head/;
		$$root[$_][-1] =~ s/(\n*)$//;
		my $lines = length $1;

		# check empty lines at start next para
		if (ref $$root[$_+1]) { # unwrap tree
			my $next = $$root[$_+1];
			next if $$next[0] =~ /^head/;
			while (ref $$next[2]) { $next = $$next[2] }
			$$next[2] =~ s/^(\n*)//;
			$lines += length $1;
			splice @$next, 2, 1 unless length $$next[2];
			$$root[$_+1] = undef unless @{$$root[$_+1]} > 2;
		}
		elsif (defined $$root[$_+1]) {
			$$root[$_+1] =~ s/^(\n*)//;
			$lines += length $1;
			$$root[$_+1] = undef unless length $$root[$_+1];
		}
		
		# Set end of line on previous para
		if ($_ > 0) {
			if (ref $$root[$_-1]) { # unwrap tree
				my $prev = $$root[$_-1];
				next if $$prev[0]  =~ /^head/;
				while (ref $$prev[-1]) { $prev = $$prev[-1] }
				$$prev[-1] =~ s/\n?$/\n/;
			}
			elsif (defined $$root[$_-1]) {
				$$root[$_-1] =~ s/\n?$/\n/;
			}
		}
		
		$$root[$_][1]{empty_lines} = $lines - 1 if $lines > 1;
	}
	@$root = grep defined($_), @$root;
		
	unshift @$root, 'Document', {};
	#use Data::Dumper; print STDERR Dumper $root;
	return $root;
}

=item C<create_link_tag(LINK_DATA)>

Returns an anonymous L<Gtk2::TextIter> that holds LINK_DATA.

Every link tag should only be applied to one text region.

=cut

sub create_link_tag {
	my($self, $data) = @_;
	
	my $tag = $self->create_tag(undef, @{$TAGS{link}});
	$tag->{is_link}   = 1;
	$tag->{link_data} = $data;

	return $tag;
}

=item C<auto_selection(TAG, ITER)>

Returns start and end iterators for the current selection.
If there is no selection in the buffer TAG and ITER are used to
automaticly select a range if possible.
ITER is optional and defaults to the cursor.

=cut

sub auto_selection {
	my ($self, $tag, $iter) = @_;
	my ($start, $end) = $self->get_selection_bounds;
	#warn "selection: $start, $end tag: $tag\n";
	if (!$end or $start == $end) { # autoselect word if defined tag
		if ($tag eq 'link') {
			($start, $end) = $self->select_link($iter);
			($start, $end) = $self->select_word($iter)
				if !$end or $start == $end;
		}
		elsif ($tag =~ /^head/) {
			($start, $end) = $self->select_line($iter);
		}
		elsif ($tag and $tag !~ /^[A-Z]/) {
			($start, $end) = $self->select_word($iter);
		}
		#warn "   => $start, $end\n";
		return undef if !$end or $start == $end;
	}
	return ($start, $end);
}

=item C<select_word(ITER)>

Selects the current word if ITER is inside a word.
Cursor is used when ITER is undefined.
Returns Gtk2::TextIter for start and end or undef.

=cut

sub select_word {
	my ($self, $iter) = @_;
	$iter ||= $self->get_iter_at_mark( $self->get_insert );
	#warn "select-word: ", $iter->inside_word, "\n";
	return unless $iter->inside_word
		or $iter->starts_word or $iter->ends_word;
	my $begin = $iter->copy;
	$begin->backward_word_start unless $begin->starts_word;
	$iter->forward_word_end unless $iter->ends_word;
	$self->select_range($begin, $iter);
	return ($begin, $iter);
}

=item C<select_line(ITER)>

Selects the line that contains ITER.
Cursor is used when ITER is undefined.
Returns Gtk2::TextIter for start and end or undef.

=cut

sub select_line {
	my ($self, $iter) = @_;
	$iter ||= $self->get_iter_at_mark( $self->get_insert );
	my $begin = $iter->copy;
	$begin = $self->get_iter_at_line( $begin->get_line )
		unless $begin->starts_line;
	while (! $iter->ends_line) { $iter->forward_char }
	$self->select_range($begin, $iter);
	return ($begin, $iter);
}

=item C<select_tag(ITER, TAG)>

Selects TAG if ITER has TAG.
Cursor is used when ITER is undefined.
Tag can be either a Gtk2::TextTag or a tag name.
Returns Gtk2::TextIter for start and end or undef.

=cut

sub select_tag {
	my ($self, $iter, $tag) = @_;
	#warn "select_tag: $tag\n";
	$tag = $self->get_tag_table->lookup($tag) unless ref $tag;
	$iter ||= $self->get_iter_at_mark( $self->get_insert );
	return unless $iter->has_tag($tag) or $iter->toggles_tag($tag);
	my $begin = $iter->copy;
	$begin->backward_to_tag_toggle($tag) unless $begin->begins_tag($tag);
	$iter->forward_to_tag_toggle($tag) unless $begin->ends_tag($tag);
	$self->select_range($begin, $iter);
	return ($begin, $iter);
}

=item C<select_link(ITER)>

Selects link if ITER is on a link.
Cursor is used when ITER is undefined.
Returns Gtk2::TextIter for start and end or undef.

=cut

sub select_link {
	#warn "select_link\n";
	my ($self, $iter) = @_;
	$iter ||= $self->get_iter_at_mark( $self->get_insert );
	for ($iter->get_tags) {
		next unless $_->{is_link};
		return $self->select_tag($iter, $_);
	}
}

=item C<apply_tag_to_selection>

TODO

=cut

sub apply_tag_to_selection {
	my ($self, $tag) = @_;
	
}

=item C<replace_selection(TEXT)>

Replace any selected text with TEXT.
Returns iterators for the start and end of the newly inserted selection.

=cut

sub replace_selection {
	my ($self, $text) = @_;
	my ($start, $end) = $self->get_selection_bounds;
	return if !$end or $start == $end;

	my $_start = $start->get_offset;
	my $_end   = $start->get_offset + length $text;

	$self->delete($start, $end);
	$self->insert($start, $text);

	($start, $end) = map $self->get_iter_at_offset($_), ($_start, $_end);
	$self->select_range($start, $end);

	return $start, $end;
}

=item C<set_edit_mode_tags>

TODO

=cut

sub set_edit_mode_tags {
	my $self = shift;
	my $table = $self->get_tag_table;
	$self->{edit_mode_tags} =
		[ map { ref($_) ? $_ : $table->lookup($_) } @$_ ];
}

=item C<set_edit_mode_from_cursor>

TODO

=cut

sub _get_open_tags {
	# iter->get_tags has "left" gravity
	# this function has "right" gravity
	my ($self, $iter) = @_;
	my @tags = $iter->get_tags;
	for my $t ($iter->get_toggled_tags(1)) { # tags that are toggles on here
		@tags = grep {$_ ne $t} @tags;
	}
	push @tags, $iter->get_toggled_tags(0);
	return @tags;
}

sub set_edit_mode_from_cursor {
	my $self = shift;
	my $iter = $self->get_iter_at_mark( $self->get_insert );
	$self->{edit_mode_tags} = [$iter->get_tags];
	#print join ' ', "Mode:", map(
	#	{$_->get_property('name')} @{$self->{edit_mode_tags}}
	#), "\n";
}

=item C<get_edit_mode_tags>

TODO

=cut

sub get_edit_mode_tags {
	$_[0]->set_edit_mode_from_cursor
		unless defined $_[0]->{edit_mode_tags};
	@{$_[0]->{edit_mode_tags}};
}

=item C<toggle_edit_mode_tag>

TODO

=cut

sub toggle_edit_mode_tag {
	my ($self, $tag) = @_;
	$tag = $self->get_tag_table->lookup($tag) unless ref $tag;
	$self->set_edit_mode_from_cursor
		unless defined $self->{edit_mode_tags};
	if (grep {$_ eq $tag} @{$self->{edit_mode_tags}}) {
		@{$self->{edit_mode_tags}} = grep {$_ ne $tag} @{$self->{edit_mode_tags}};
	}
	else {
		push @{$self->{edit_mode_tags}}, $tag;
	}
}

=item C<reset_edit_mode>

TODO

=cut

sub reset_edit_mode { $_[0]->{edit_mode_tags} = undef }

=item C<undo>

TODO

=cut

sub undo {
}

=item C<redo>

TODO

=cut

sub redo {
}

=item C<copy_clipboard>

TODO

=cut

sub copy_clipboard {
	my $self = shift;
	$self->begin_user_action;
	$self->_cut_or_copy(shift, 0, 1, shift);
	$self->end_user_action;
}

=item C<cut_clipboard>

TODO

=cut

sub cut_clipboard {
	my $self = shift;
	$self->begin_user_action;
	$self->_cut_or_copy(shift, 1, 1, shift);
	$self->end_user_action;
}

sub _cut_or_copy {
	my ($self, $clipboard, $delete, $interactive, $editable) = @_;
	my ($start, $end) = $self->get_selection_bounds;
	
	# TODO check anchor
	
	return if !$start or $start->equal($end);

	my $tree = $self->get_parse_tree($start, $end);
	$tree->[1]{comment} = 'Copy-paste buffer';
	
	$clipboard->set_with_data(
		\&_clipboard_get, \&_clipboard_clear, $tree, @_targets)
		or return;

	#$interactive ? $self->delete_interactive($start, $end, $editable) :
	$self->delete($start, $end)
		if $delete ;
}

sub _clipboard_get { # clipboard callback
	my ($clipboard, $selection, $id, $tree) = @_;
	#print STDERR '_clipboard_get type: '.$selection->target->name."\n";
	
	if ($selection->target->name eq 'ZIM_TREE') {
		$_clipboard = $tree;
		#print STDERR "_clipboard_get set ref: $tree\n";
		$selection->set(
			Gtk2::Gdk::Atom->intern('ZIM_TREE', 0),
			8, "$tree" );
	}
	else { # text
		my $string = _dump_tree($tree);
		#print STDERR "_clipboard_get set_text >>$string<<\n";
		$selection->set_text($string);
	}
}

sub _clipboard_clear {
	my $tree = pop;
	$_clipboard = undef if "$tree" eq "$_clipboard";
	@$tree = ();
}

sub _dump_tree { # TODO add "\n" around headings and support lists
	# Make sure not to modify content of the reference !
	my $tree = shift;
	my $string = '';
	$string .= ref($_) ? _dump_tree($_) : $_
		for @$tree[2 .. $#$tree];
	return $string;
}

=item C<paste_clipboard>

TODO

=cut

sub paste_clipboard {
	my ($self, $clipboard, $insert, $editable) = @_;
	
	# check selections
	my $replace = 0;
	my ($start, $end) = $self->get_selection_bounds;
	if (defined $start) {
		my $iter = $insert || $self->get_iter_at_mark( $self->get_insert );
		if ($iter->in_range($start, $end) or $iter->equal($end)) {
			$insert = $iter;
			$replace = 1;
		}
	}	
	
	# check insert point
	$self->create_mark('gtk_paste_point_override', $insert, 0)
		if defined $insert ;
	
	my $paste_data = {
		buffer      => $self,
		interactive => 1,
		editable    => $editable,
		replace     => $replace,
	};
	
	$clipboard->request_contents(
		Gtk2::Gdk::Atom->intern('ZIM_TREE', 0),
		\&_clipboard_paste, $paste_data   );
}

sub _clipboard_paste {
	# Paste data type negotiating
	my ($clipboard, $selection, $paste_data) = @_;
	my $type = $selection->target->name;
	my $data = $selection->data;
	#print STDERR '_clipboard_paste type: '.$selection->target->name.", data: $data\n";
	
	if ($type eq 'ZIM_TREE' and $_clipboard and $data eq "$_clipboard") {
		$paste_data->{type} = $type;
		&_clipboard_paste_data($clipboard, $data, $paste_data);
	}
	else {
		# Not pasting from a zim widget 
		# or pasting from a different process
		$paste_data->{type} = 'STRING';
		$clipboard->request_text(\&_clipboard_paste_data, $paste_data);
	}
}

sub _clipboard_paste_data {
	# Actual paste
	my ($clipboard, $string, $paste_data) = @_;
	my $self = $paste_data->{buffer};
	my $type = $paste_data->{type};
	#print STDERR "_clipboard_paste_data type: $type, string: $string\n";

	return unless length $string;
	return if $type eq 'ZIM_TREE' and "$string" ne "$_clipboard";

	$self->begin_user_action if $paste_data->{interactive};
	
	# check selection
	$self->delete_selection(@{$paste_data}{'interactive','editable'})
		if $paste_data->{replace};
	
	# check gtk_paste_point_override
	my $iter;
	my $mark = $self->get_mark('gtk_paste_point_override');
	if ($mark) {
		$iter = $self->get_iter_at_mark( $mark );
		$self->delete_mark($mark);
	}
	else { $iter = $self->get_iter_at_mark( $self->get_insert ) }
	
	# actual insert
	if ($paste_data->{type} eq 'ZIM_TREE') {
		#use Data::Dumper; print STDERR '_clipboard: ', Dumper $_clipboard;
		$self->insert_blocks($iter, $_clipboard);
	}
	else { # text
		$paste_data->{interactive}
			? $self->insert_interactive(
				$iter, $string, $paste_data->{editable} )
			: $self->insert($iter, $string) ;
	}

	$self->end_user_action if $paste_data->{interactive};
}

1;

__END__

=back

=head1 TAGS

Tags are used in L<Gtk2::TextBuffer> to render certain parts of the text buffer with
a certain colours or certain decorations. In B<Gtk2::Ex::HyperTextBuffer> they also
represent a node in the document parse tree. See L<Gtk2::TextTag> for more details.

The following tags are pre-defined:

=over 4

=item head1

=item head2

=item head3

=item head4

=item head5

=item bold

=item italic

=item underline

=item verbatim

=item Verbatim

Like 'verbatim' but intended for multiline blocks.

=item link

For links anonymous TextTags are used, but the visual properties are defined the
same way as for other tags.

=back

=head1 AUTHOR

Jaap Karssenberg (Pardus) E<lt>pardus@cpan.orgE<gt>

Copyright (c) 2005 Jaap G Karssenberg. All rights reserved.
This program is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 SEE ALSO

=cut

